#!/usr/bin/env python

# Author                 Version      Date
# -----------------------------------------------
# S M Crawford (SAAO)    0.3          10 Apr 2008

# saltcrclean will clean multi-extension data of cosmic rays 

from pyraf import iraf 

import saltprint, saltkey, saltio, saltstring, saltstat, salttran
import os, time
import numpy, pyfits


# -----------------------------------------------------------
# core routine

def saltcrclean(images,outimages,outpref,crtype,thresh,mbox, flux_ratio, bbox,   \
                bthresh, gain, rdnoise, fthresh, bfactor, gbox, gthresh, maxiter,\
                clobber,logfile,verbose,status):

    status = 0

# test the logfile

    logfile = saltio.logname(logfile)

# log the call 

    saltprint.line(logfile,verbose)
    history = 'SALTCRCLEAN  -- '
    history += 'images=%s ' % images
    history += 'outimages=%s ' % outimages
    history1 = 'outpref=%s ' % outpref
    history1 += 'crtype=%s ' % crtype
    history1 += 'thresh=%3.2f '%thresh
    history1 += 'mbox=%i ' % mbox  
    history1 += 'flux_ratio=%i ' % mbox  
    history1 += 'bbox=%i ' % bbox
    history1 += 'bthresh=%3.2f '%bthresh
    history1 += 'gain=%3.2f ' % gain
    history1 += 'rdnoise=%3.2f ' % rdnoise
    history1 += 'fthresh=%3.2f  '% fthresh
    history1 += 'bfactor=%i ' % bfactor
    history1 += 'gbox=%i ' % gbox
    history1 += 'gthresh=%3.2f '%thresh
    history1 += 'maxiter=%i ' % maxiter
    history1 += 'logfile=%s ' % logfile
    yn = 'n'
    if (clobber): yn = 'y'
    history1 += 'clobber=%s ' % yn
    yn = 'n'
    if (verbose): yn = 'y'
    history1 += 'verbose=%s ' % yn
    history = history + history1
    saltprint.log(logfile,history,verbose)

# start time

    saltprint.time('\nSALTCRCLEAN  -- started at',logfile,verbose)
    saltprint.log(logfile,' ',verbose)

# are the arguments defined
   
    if (status == 0): status = saltio.argdefined('images',images,logfile)
    if (status == 0): status = saltio.argdefined('logfile',logfile,logfile)

# if the input file is a list, does it exist?

    if (status == 0 and images[0] == '@'): 
        status = saltio.listexists('Input',images,logfile)

# parse list of input files

    if (status == 0):
        infiles, status = saltio.listparse('Raw image',images,'','','',logfile)

# check input files exist

    if (status == 0): status = saltio.filesexist(infiles,'','r',logfile)

# load output name list: @list, * and comma separated
   
    outimages = outimages.strip()
    outpref = outpref.strip()
    if (status == 0 and len(outpref) == 0 and len(outimages) == 0):
        message = '\nERROR -- SALTCRCLEAN : output file(s) not specified'
        status = saltprint.err(logfile,message)

# test output @filelist exists

    if (status == 0 and len(outimages) > 0 and outimages[0] == '@'): 
        status = saltio.listexists('Output',outimages,logfile)

# parse list of output files

    if (status == 0):
        outfiles, status = saltio.listparse('Output image',outimages,outpref,infiles,'',logfile)

# are input and output lists the same length?

    if (status == 0): status = saltio.comparelists(infiles,outfiles,'Input','output',logfile)

# do the output files already exist?

    if (status == 0 and not clobber): status = saltio.filesexist(outfiles,'','w',logfile)


# Begin processes each file 

    ifil = 0
    if status == 0:
     for infile in infiles:
	 pstatus=status

	 #determine the name for the output file
	 outfile = outfiles[ifil]
	 outpath = outfile.rstrip(os.path.basename(outfile))
	 if (len(outpath) == 0): outpath = '.'

	 #open each raw image
	 if (pstatus == 0):
	    struct, pstatus = saltio.openfits(infile,logfile)

	 # how many extensions?

	 if (pstatus == 0): 
	    try:
		nextend=len(struct)-1
		hdustart=1
		if nextend==0: 
		   nextend=1
		   hdustart=0
	    except:
		message='SALTCRCLEAN--ERROR:  No extensions in %s to process' % infile
                pstatus = saltprint.err(logfile,message)

	 # CCD on-chip binning

	 #if (pstatus == 0): xbin, ybin, pstatus = saltkey.ccdbin(struct[0],infile,logfile)

	 # create output hdu structure

	 if (pstatus == 0): 
	    outstruct = [None] * (nextend+hdustart)
	    outstruct[0]=struct[0]

	 # if the image only has a single extension
	 if hdustart==0:
	    if (pstatus==0):
	       outstruct[0].data, pstatus= cleancosmicrays(struct[0].data, crtype, thresh, mbox, bbox, bthresh, flux_ratio, \
                                                  gain, rdnoise, bfactor, fthresh, gbox, gthresh, maxiter, logfile)

	 #if the image has a primary and multiple extensions
	 if hdustart==1:
	  for i in range(nextend):
	      #Add the outimage to the output structure
	      if (pstatus==0):
		 hdu=i+hdustart
		 outarr, pstatus=cleancosmicrays(struct[hdu].data, crtype, thresh, mbox, bbox, bthresh, flux_ratio, \
                                        gain, rdnoise, bfactor, fthresh, gbox, gthresh, maxiter, logfile)
		 outstruct[hdu] = pyfits.ImageHDU(outarr)
		 outstruct[hdu].header=struct[hdu].header
		

         # close input FITS file

	 if (pstatus == 0): pstatus = saltio.closefits(struct,logfile)

         # housekeeping keywords


	 if (pstatus == 0): pstatus = saltkey.put('SAL-TLM',time.asctime(time.localtime()),
					       outstruct[0],outfile,logfile)
	 if (pstatus == 0): 
                history = 'SALTCRCLEAN  -- '
                history += 'images=%s ' % infile
                history += 'outimages=%s ' % outfile
		history += history1
		pstatus = saltkey.history(outstruct[0],history,outfile,logfile)


	 # clobber old files
	    
	 if (pstatus == 0 and clobber and os.path.isfile(outfile)): 
	    pstatus = saltio.delete(outfile,False,logfile)

	 # write FITS file of mosaiced image

	 if (pstatus == 0) and hdustart==1: 
                 outstruct = pyfits.HDUList(outstruct)
                 outstruct.writeto(outfile)
	 if (pstatus == 0) and hdustart==0: 
		 pyfits.writeto(outfile, outstruct[0].data, outstruct[0].header)
	 if (pstatus == 0): 
	    message = 'SALTCRCLEAN  -- created ' + os.path.basename(outfile) + ' at'
	    saltprint.time(message,logfile,verbose)
	 ifil += 1

# end time

    if (status == 0):
	saltprint.time('SALTCRCLEAN  -- completed at',logfile,verbose)
    else:
	saltprint.time('\nSALTCRCLEAN  -- aborted at',logfile,verbose)

#-----------------------------------------------------------
# Clean cosmic rays

def cleancosmicrays(arr, crtype, thresh, mbox, bbox, bthresh, flux_ratio, gain, \
                    rdnoise, b_factor, fthresh, gbin, gthresh, max_iter, logfile):
    """Clean the cosmic rays from an input array using three different types of methods
    """
    status=0

    #makesure the data array is in floating point
    try:
	arr=arr*1.0
    except:
	message='SALTCRCLEAN--ERROR:Array cannot be convert to the correct data type'
	saltio.err(logfile, message)
	return arr

    #apply the correct function to idenitfy cosmic rays

    if crtype=='median':
	arr, crarr, status=crmedian(arr, thresh, mbox, bbox, bthresh, gbin, gthresh, max_iter, logfile)
    elif crtype=='fast':
	arr, crarr, status=crfast(arr, thresh, mbox, bthresh, flux_ratio, gbin, gthresh,max_iter, logfile)
    elif crtype=='edge':
	arr, crarr, status=credge(arr, thresh, mbox, gain, rdnoise, b_factor, fthresh, gbin, gthresh, max_iter, logfile)
    else:
	message='SALTCRCLEAN--ERROR:  CRTYPE %s is not a valid value' % crtype
	status=saltprint.err(logfile, message)


    return arr, status

#-----------------------------------------------------------
# set up a boxes dimensions

def setbox(x, y, mbox, xlen, ylen):
	#set up the edges of the array
	y1=y-mbox
	y2=y+mbox+1
	x1=x-mbox
	x2=x+mbox+1

	#check to make the data are inside the array
	if y1 < 0:  y1=0
	if x1 < 0:  x1=0
	if y2 >= ylen: y2=ylen-1
	if x2 >= xlen: x2=xlen-1

	return x1, x2, y1, y2


#-----------------------------------------------------------
# Fast identification of a cosmic ray

def idcray(arr, flux_ratio, y, x, ylen, xlen, mbox, logfile):

	status=0
	#set up the box dimensions
	x1, x2, y1, y2 = setbox(x, y, mbox, xlen, ylen)

	#is the source the brightest in the box
	if arr[y,x] < arr[y1:y2,x1:x2].max(): 
	     status=1
	     return y, x, 0, status

	#If the pixel to median flux  is less than 
	# the flux ratio, reject the objects as a comsic ray
	median_pix, status=saltstat.median2d(arr[y1:y2,x1:x2], logfile)
	if abs(arr[y,x]/median_pix) < flux_ratio:
	     status=1
	     return y, x, 0, status

	return y,x,median_pix,status
	


#-----------------------------------------------------------
# Fast clean of the cosmic rays

def crfast(arr, thresh, mbox, bthresh, flux_ratio, gbin, gthresh, max_iter, logfile):
	#import imagestats

	#set up the size of the box.  Minimimum size is a 3x3 box around
        # the central pixel
	if mbox < 3: mbox=3
	mbox=int(mbox/2)
	
	status=0
	onpix=0
	npix=1
	niter=0
	ylen=len(arr)
	xlen=len(arr[0])
	oarr=arr.copy()
	
	#calculate the background statistics
	#the imstat's are commented out as they might not be a standard package
	#imstat=imagestats.ImageStats(arr, nclip=max_iter)
	#mean=imstat.mean
	#bsigma=imstat.stddev
	mean, midpt, bsigma=saltstat.iterstat(arr, bthresh, max_iter) 
	thresh=bsigma*thresh

	#set up the cr identification array
	crarr=0.0*arr

	while niter < max_iter and npix>onpix:
	 #Detect all possible objects on the image
	 rarr=arr-mean 
	 darr= numpy.where(rarr> thresh)

	 #Out of those detections, eliminate those that aren't cosmic rays
	 for i in range(len(darr[0])):
	  y=darr[0][i]
	  x=darr[1][i]
	  y, x, mpix, cstatus=idcray(rarr, flux_ratio, y, x, ylen, xlen, mbox, logfile)
	  if cstatus==0: crarr[y,x]= mpix+mean

	 #correct the image
	 mask=1*(crarr!=0)
	 arr=arr*(1-mask)+crarr*mask

	 #increment the counter
	 onpix=npix
	 npix=mask.sum()
	 niter +=1

	#grow the sources
	if gbin > 0:
	  if gbin < 3: gbin=3
	  gbin=int(gbin/2)
	  gthresh =bsigma*gthresh
	  garr=1.0*(rarr>gthresh)
	
	  #the first step is to grow the masks
	  darr=numpy.where(mask>0)
	  for i in range(len(darr[0])):
	   y=darr[0][i]
	   x=darr[1][i]
	   x1, x2, y1, y2 = setbox(x, y, gbin, xlen, ylen)
	   crarr[y1:y2,x1:x2]=garr[y1:y2,x1:x2]

	  #second step is to pass back though and replace 
	  #each pixel with the median of the pixels around it
	  mask=1*(crarr!=0)
	  darr=numpy.where(mask>0)
          for i in range(len(darr[0])):
           y=darr[0][i]
           x=darr[1][i]
           x1, x2, y1, y2 = setbox(x, y, gbin, xlen, ylen)
	   crarr[y,x], status=saltstat.median2d(arr[y1:y2,x1:x2], logfile)
	  arr=arr*(1-mask)+crarr*mask

	#correct the image 
	crarr=1.0*((oarr-arr)>0)
	return arr, crarr, status

#-----------------------------------------------------------
# Function to calculate the sigma at each pixel

def sigma_func (arr):
    std=1
    arr=arr.ravel()
    arr=numpy.sort(arr)
    larr=len(arr)
    x1=int(0.159*larr)
    x2=int(0.841*larr)
    std=(arr[x2]-arr[x1])/2.0
    return std



#-----------------------------------------------------------
# Median clean the cosmic rays

def crmedian(arr, thresh, mbox, bbox, bthresh, gbin, gthresh, max_iter, logfile):
    """Identify cosmic rays through median technique
    """
    from scipy.signal import fftconvolve as conv2d
    from scipy.ndimage.filters import generic_filter
    status=0
    niter=0
    npix=1
    oarr=arr.copy()

    #Check to make sure the background box is an appropriate size
    if bbox <  10:
	bbox=10
	message='SALTCRCLEAN--WARNING:  Background box size < 10, setting to 10'
	saltprint.err(logfile, message)

    while niter < max_iter and npix:
     #make the background image
     barr=generic_filter(arr,sigma_func, size=(bbox,bbox))

     #Median smooth the image
     marr=saltstat.median_image(arr, mbox, logfile)
 
     #Find the residual image
     rarr=(arr-marr)/barr

     #identify all sources
     crarr=1.0*(rarr > thresh)

     #grow the sources
     if gbin>0:
	if gbin<3: gbin=3
        g_conv = numpy.zeros([gbin,gbin])+1.0
	maskg =  conv2d(crarr,g_conv,mode='same')
        maskg =  (maskg >= 1)
        maskg =  1.0*(arr*maskg > gthresh)
        crarr = crarr + maskg
        crarr = (crarr >= 1) * 1.0

     #replace the pixels
     arr = arr * (1-crarr)+marr*crarr

     #increment the counter
     niter += 1
     npix=crarr.sum()


    crarr=1.0*((oarr-arr)>0)
    arr=oarr
    return arr, crarr, status

#-----------------------------------------------------------
# Edge clean the cosmic rays

def credge(arr, thresh, mbox, gain, rdnoise, b_factor, fthresh, gbin, gthresh, max_iter, logfile):
    """Identify cosmic rays following the van Dokkem (2001) prescription
    """
    from scipy.signal import fftconvolve as conv2d

    status=0
    min_limit=0.01

    niter=0
    npix= 1


    #set up the convolution kernel
    f_conv=numpy.zeros([3,3])
    f_conv[0][0]=0
    f_conv[0][1]=-1
    f_conv[0][2]=0
    f_conv[1][0]=-1
    f_conv[1][1]=4
    f_conv[1][2]=-1
    f_conv[2][0]=0
    f_conv[2][1]=-1
    f_conv[2][2]=0

    if gbin > 0: 
	if gbin<3: gbin=3
	g_conv = numpy.zeros([gbin,gbin])+1.0

    #set up the array
    shape=arr.shape
    sdata=arr

    #iterate until all the cosmic rays are detected or max_iterations are hit
    while npix > 0 and niter < max_iter and status==0:
         #rebin the data
         newshape=(b_factor*shape[0],b_factor*shape[1])
         ldata,status =salttran.rebin_factor(sdata,newshape, logfile)
         #convolve with f_conv
	 if status==0:
	    try:
               ldata=conv2d(ldata,f_conv,mode='same')
               mask = (ldata >= 0)
               ldata = ldata * mask
	    except:
	       message='SALTCRCLEAN--ERROR:  Unable to convole the laplace filter with the image'
	       status=saltprint.err(logfile, message)\

	 #return to the original binning
	 if status==0:
            ldata, status = salttran.blockave(ldata,b_factor,newshape, logfile)
         
         #create noise model
         if status==0:
	    meddata=saltstat.median_image(sdata,5, logfile)
            noise = abs(meddata)
            noise = numpy.sqrt(gain*noise+rdnoise**2)/gain

         #create S/N image
	 if status==0:
            odata = ldata / noise / b_factor

         #remove extended objects
	 if status==0:
            odata = odata - saltstat.median_image(odata,5, logfile)


         #select objects
	 if status==0:
           masks = (odata>thresh)*1.0

         #remove compact bright sources
	 if status==0:
            mdata = saltstat.median_image(sdata,3, logfile)
            fdata = mdata - saltstat.median_image(mdata,7, logfile)
            fdata = fdata / noise
	    # set a minimum value for all pixels so no divide by zero problems
	    indf = numpy.where(fdata < min_limit)
	    fdata[indf]=min_limit

            fdata = odata * masks/fdata
            maskf = (fdata > fthresh)

         #make the list of cosmic rays
	 if status==0:
            mask =  masks*maskf

	 #grow the cosmic rays
	 if gbin > 0:
	    maskg =  conv2d(mask,g_conv,mode='same')
            maskg =  (maskg >= 1)
            maskg =  1.0*(odata*maskg > gthresh)
            mask = mask + maskg
            mask = (mask >= 1) * 1.0

	 #replace the cosmic rays
	 sdata = sdata * (1-mask)+meddata*mask

	 #increment the counter
	 npix=(mask > 0).sum()
	 niter += 1

    crarr=1.0*((arr-sdata)>0)
    arr=sdata
    return arr, crarr, status



#-----------------------------------------------------------
# housekeeping keywords

def addhousekeeping(struct, outhdu, outfile, logfile):
    status=0
    if (status == 0): status = saltkey.put('SAL-TLM',time.asctime(time.localtime()), \
				       struct,outfile,logfile)
    if (status == 0): status = saltkey.new('EXTNAME','SCI','Extension name',         \
				       struct,outfile,logfile)
    if (status == 0): status = saltkey.new('EXTVER',outhdu,'Extension number',       \
				       struct,outfile,logfile)

    return struct, status

# -----------------------------------------------------------
# main code 

parfile = iraf.osfn("saltred$saltcrclean.par") 
t = iraf.IrafTaskFactory(taskname="saltcrclean",value=parfile,function=saltcrclean, pkgname='saltred')
