#!/usr/bin/env python

# Author                 Version      Date
# -----------------------------------------------
# Martin Still (SAAO)    1.0          25 Jul 2007

# saltlog reads the critical header keywords of SALT FITS data and
# collates them into a FITS table. This can be used for the basis
# of an observation log or as a meta-table for pipeline processing.

from pyraf import iraf 
import saltprint, saltio, saltkey
import os, glob, time, pyfits

# -----------------------------------------------------------
# core routine

def saltlog(images,outfile,clobber,logfile,verbose,status):

    filename = []  # file name
    instrume = []  # detector insturment
    propid = []    # propsoal ID
    proposer = []  # PI name
    detswv = []    # instrument software version
    object = []    # object name
    ra = []        # RA coordinate
    dec = []       # Dec coordinate
    pm_ra = []     # RA proper motion (arcsec)
    pm_dec = []    # Dec proper motion (arcsec)
    objepoch = []  # epoch of ra and dec above
    equinox =  []  # equnix of target coordinates
    date_obs = []  # date of observation
    utc_obs = []   # UT start of observation
    time_obs = []  # UT start of observation
    exptime = []   # exposure time
    filpos = []    # filter position
    detsize = []   # detector size (raw pixels)
    nccds = []     # number of CCDs in detector
    ccdsum = []    # CCD binning
    gainset = []   # CCD gain setting
    rospeed = []   # CCD readout speed
    ccdtem = []    # CCD temperature (K)
    dewtem = []    # cold box temperature (C)
    amptem = []    # SDSU controller temperature (C)
    centem = []    # cold end temperature (C)
    dewpre = []    # dewar pessure (units unspecified)
    pcs_ver = []   # version of RSS control software
    posang = []    # planned taget position angle (units unspecified)
    ccdtype = []   # CCD type
    lampid = []    # name of lamp
    calfilt = []   # name of calibration filter
    calnd = []     # calibration neutral density (percentage?)
    trkx = []      # tracker x
    trky = []      # tracker y
    trkz = []      # tracker z
    trkphi = []    # tracker phi
    trktheta = []  # tracker theta
    trkrho = []    # tracker rho
    intfr = []     # interferometer reading
    colphi = []    # autocollimator phi (degrees)
    coltheta = []  # autocollimator theta (degrees)
    telha = []     # hour angle of telescope
    telra = []     # telescope RA
    teldec = []    # telescope Dec
    telpa = []     # telescope position angle
    telrho = []    # position angle of rho stage at readout (degrees)
    telaz = []     # telescope azimuth at readout (degrees)
    telalt = []    # telescope evelvation at readout
    teltemp = []   # telescope temperature (C)
    payltemp = []  # payload temperature (C)
    pellicle = []  # position of SALTICAM pellicle
    instport = []  # instrument port
    cf_state = []  # config state machine state
    obsmode = []   # config state machine state
    sm_state = []  # slitmask machine state
    sm_sta = []    # commanded slitmask station
    sm_steps = []  # slitmask magazine position (steps)
    sm_volts = []  # slitmask analog encoder (volts)
    sm_sta_s = []  # slitmask station from counting steps
    sm_sta_v = []  # slitmask station from the analog encoder
    maskid = []    # slitmask bar code
    masktyp = []   # slitmask mask type
    wp_state = []  # waveplate state machine state
    hwp_cmd = []   # commanded half wave station
    hw_steps = []  # half wave position (steps)
    hwp_sta = []   # half wave plate position station from incremental encoder
    qwp_cmd = []   # commanded quarter wave station
    qw_steps = []  # quarter wave position (steps)
    qwp_sta = []   # quarter wave plate position station from incremental encoder
    qwp_ang = []   # commanded quarter wave angle
    hwp_ang = []   # commanded half wave angle
    sh_state = []  # shutter state machine state
    fo_state = []  # focus state machine state
    fo_pos = []    # commanded focus position
    focus = []     # focus position (steps)
    fo_volts = []  # focus position from analog encoder (volts)
    fo_pos_s = []  # focus position from step count
    fo_pos_v = []  # focus position from analog encoder
    gr_state = []  # grating state machine state
    gr_sta = []    # commanded grating station
    gr_angle = []  # grating rotation angle (steps)
    gm_steps = []  # grating magazine position (steps)
    gm_volts = []  # grating magazine position (encoder)
    gr_sta_s = []  # grating station from steps
    gr_sta_v = []  # grating station from volts
    gr_steps = []  # grating rotation angle (steps)
    grating = []   # commanded grating station
    grtilt = []    # commanded grating angle
    bs_state = []  # beamsplitter state machine state
    fi_state = []  # filter state machine state
    fi_sta = []    # commanded filter station
    fm_steps = []  # filter magazone position (steps)
    fm_volts = []  # filter magazone position from analog encoder
    fm_sta_s = []  # filter magazine position from steps
    fm_sta_v = []  # filter magazine position from analog encoder
    filter = []    # filter bar code
    ar_state = []  # articulation state machine state
    ar_sta = []    # commanded articulation station
    camang = []    # commanded articulation station
    ar_sta_s = []  # articulation station from incremental order
    ar_angle = []  # articulation angle from incremental order
    coltemp = []   # collimator temperature (C)
    camtemp = []   # camera temperture (C)
    proc = []      # PFIS temperature
    detmode = []   # detector mode
    col = []       # FITS table columns

# test the logfile

    logfile = saltio.logname(logfile)

# log the call 

    saltprint.line(logfile,verbose)
    message = 'SALTLOG -- '
    message += 'images='+images+' '
    message += 'outfile='+outfile+' '
    yn = 'n'
    if (clobber): yn = 'y'
    message += 'clobber='+yn+' '
    message += 'logfile='+logfile+' '
    yn = 'n'
    if (verbose): yn = 'y'
    message += 'verbose='+yn
    saltprint.log(logfile,message,verbose)
    history1 = message

# start time

    saltprint.time('\nSALTLOG -- started at',logfile,verbose)
    saltprint.log(logfile,' ',verbose)

# is the input file specified?
   
    if (status == 0): status = saltio.filedefined('Input',images,logfile)

# if the input file is a list, does it exist?

    if (status == 0 and images[0] == '@'): 
        status = saltio.listexists('Input',images,logfile)

# parse list of input files

    if (status == 0):
        infiles, status = saltio.listparse('Raw image',images,'','','',logfile)

# check input files exist

    if (status == 0): status = saltio.filesexist(infiles,'','r',logfile)

# does the output file already exist?

    if (status == 0 and not clobber): 
	status = saltio.filedoesnotexist(outfile,verbose,logfile)

# interate over and open image files

    for infile in infiles:
	if (status == 0): struct,status = saltio.openfits(infile,logfile)

# instrument

	scam = False
	rss = False
	if (status == 0): 
	    instrume = appendkey(instrume,'INSTRUME','',struct[0],infile,logfile)
	    if (instrume[len(instrume)-1] == 'RSS'): rss = True 
	    if (instrume[len(instrume)-1] == 'SALTICAM'): scam = True 

# ingest primary keywords from files in the image list

	if (status == 0): 
	    filename.append(infile.split('/')[len(infile.split('/'))-1])
	    propid = appendkey(propid,'PROPID','UNKNOWN',struct[0],infile,logfile)
	    proposer = appendkey(proposer,'PROPOSER','UNKNOWN',struct[0],infile,logfile)
	    obsmode = appendkey(obsmode,'OBSMODE','UNKNOWN',struct[0],infile,logfile)
	    detmode = appendkey(detmode,'DETMODE','UNKNOWN',struct[0],infile,logfile)
	    ccdtype = appendkey(ccdtype,'CCDTYPE','UNKNOWN',struct[0],infile,logfile)
	    detswv = appendkey(detswv,'DETSWV','UNKNOWN',struct[0],infile,logfile)
	    object = appendkey(object,'OBJECT','UNKNOWN',struct[0],infile,logfile)
	    ra = appendkey(ra,'RA','UNKNOWN',struct[0],infile,logfile)
	    dec = appendkey(dec,'DEC','UNKNOWN',struct[0],infile,logfile)
	    objepoch = appendkey(objepoch,'OBJEPOCH',-999.,struct[0],infile,logfile)
	    equniox = appendkey(equinox,'EQUINOX',-999.,struct[0],infile,logfile)
	    date_obs = appendkey(date_obs,'DATE-OBS','UNKNOWN',struct[0],infile,logfile)
	    utc_obs = appendkey(utc_obs,'UTC-OBS','UNKNOWN',struct[0],infile,logfile)
	    time_obs = appendkey(time_obs,'TIME-OBS','UNKNOWN',struct[0],infile,logfile)
	    exptime = appendkey(exptime,'EXPTIME',-999.,struct[0],infile,logfile)
	    detsize = appendkey(detsize,'DETSIZE','UNKNOWN',struct[0],infile,logfile)
	    filter = appendkey(filter,'FILTER','UNKNOWN',struct[0],infile,logfile)
	    nccds = appendkey(nccds,'NCCDS',-999.,struct[0],infile,logfile)
	    ccdsum = appendkey(ccdsum,'CCDSUM','UNKNOWN',struct[0],infile,logfile)
	    gainset = appendkey(gainset,'GAINSET','UNKNOWN',struct[0],infile,logfile)
	    rospeed = appendkey(rospeed,'ROSPEED','UNKNOWN',struct[0],infile,logfile)
	    focus = appendkey(focus,'FOCUS',-999,struct[0],infile,logfile)
	    telha = appendkey(telha,'TELHA',-999.,struct[0],infile,logfile)
	    telra = appendkey(telra,'TELRA',-999.,struct[0],infile,logfile)
	    teldec = appendkey(teldec,'TELDEC',-999.,struct[0],infile,logfile)
	    telpa = appendkey(telpa,'TELPA',-999.,struct[0],infile,logfile)
	    telaz = appendkey(telaz,'TELAZ',-999.,struct[0],infile,logfile)
	    telalt = appendkey(telalt,'TELALT',-999.,struct[0],infile,logfile)
	    trkx = appendkey(trkx,'TRKX',-999.,struct[0],infile,logfile)
	    trky = appendkey(trky,'TRKY',-999.,struct[0],infile,logfile)
	    trkz = appendkey(trkz,'TRKZ',-999.,struct[0],infile,logfile)
	    trkphi = appendkey(trkphi,'TRKPHI',-999.,struct[0],infile,logfile)
	    trktheta = appendkey(trktheta,'TRKTHETA',-999.,struct[0],infile,logfile)
	    trkrho = appendkey(trkrho,'TRKRHO',-999.,struct[0],infile,logfile)
	    intfr = appendkey(intfr,'INTFR',-999.,struct[0],infile,logfile)
	    colphi = appendkey(colphi,'COLPHI',-999.,struct[0],infile,logfile)
	    coltheta = appendkey(coltheta,'COLTHETA',-999.,struct[0],infile,logfile)
	    teltemp = appendkey(teltemp,'TELTEMP',-999.,struct[0],infile,logfile)
	    payltemp = appendkey(payltemp,'PAYLTEMP',-999.,struct[0],infile,logfile)
	    ccdtem = appendkey(ccdtem,'CCDTEM',-999.,struct[0],infile,logfile)
	    dewtem = appendkey(dewtem,'DEWTEM',-999.,struct[0],infile,logfile)
	    amptem = appendkey(amptem,'AMPTEM',-999.,struct[0],infile,logfile)
	    centem = appendkey(centem,'CENTEM',-999.,struct[0],infile,logfile)

# RSS-only keywords

	if (status == 0 and rss): 
	    dewpre = appendkey(dewpre,'DEWPRE',-999.,struct[0],infile,logfile)
	    pcs_ver = appendkey(pcs_ver,'PCS-VER','UNKNOWN',struct[0],infile,logfile)
	    posang = appendkey(posang,'POSANG',-999.,struct[0],infile,logfile)
	    lampid = appendkey(lampid,'LAMPID','UNKNOWN',struct[0],infile,logfile)
	    calfilt = appendkey(calfilt,'CALFILT','UNKNOWN',struct[0],infile,logfile)
	    calnd = appendkey(calnd,'CALND',-999.,struct[0],infile,logfile)
	    telrho = appendkey(telrho,'TELRHO',-999.,struct[0],infile,logfile)
	    pellicle = appendkey(pellicle,'PELLICLE','UNKNOWN',struct[0],infile,logfile)
	    instport = appendkey(instport,'INSTPORT','UNKNOWN',struct[0],infile,logfile)
	    cf_state = appendkey(cf_state,'CF-STATE','UNKNOWN',struct[0],infile,logfile)
	    sm_state = appendkey(sm_state,'SM-STATE','UNKNOWN',struct[0],infile,logfile)
	    sm_sta = appendkey(sm_sta,'SM-STA','UNKNOWN',struct[0],infile,logfile)
	    sm_steps = appendkey(sm_steps,'SM-STEPS',-999,struct[0],infile,logfile)
	    sm_volts = appendkey(sm_volts,'SM-VOLTS',-999.,struct[0],infile,logfile)
	    sm_sta_s = appendkey(sm_sta_s,'SM-STA-S',-999.,struct[0],infile,logfile)
	    sm_sta_v = appendkey(sm_sta_v,'SM-STA-V',-999.,struct[0],infile,logfile)
	    maskid = appendkey(maskid,'MASKID','UNKNOWN',struct[0],infile,logfile)
	    masktyp = appendkey(masktyp,'MASKTYP','UNKNOWN',struct[0],infile,logfile)
	    wp_state = appendkey(wp_state,'WP-STATE','UNKNOWN',struct[0],infile,logfile)
	    hwp_cmd = appendkey(hwp_cmd,'HWP-CMD','UNKNOWN',struct[0],infile,logfile)
	    hw_steps = appendkey(hw_steps,'HW-STEPS',-999,struct[0],infile,logfile)
	    hwp_sta = appendkey(hwp_sta,'HWP-STA',-999.,struct[0],infile,logfile)
	    qwp_cmd = appendkey(qwp_cmd,'QWP-CMD','UNKNOWN',struct[0],infile,logfile)
	    qw_steps = appendkey(qw_steps,'QW-STEPS',-999,struct[0],infile,logfile)
	    qwp_sta = appendkey(qwp_sta,'QWP-STA',-999.,struct[0],infile,logfile)
	    qwp_ang = appendkey(qwp_ang,'QWP-ANG',-999,struct[0],infile,logfile)
	    hwp_ang = appendkey(hwp_ang,'HWP-ANG',-999,struct[0],infile,logfile)
	    sh_state = appendkey(sh_state,'SH-STATE','UNKNOWN',struct[0],infile,logfile)
	    fo_state = appendkey(fo_state,'FO-STATE','UNKNOWN',struct[0],infile,logfile)
	    fo_pos = appendkey(fo_pos,'FO-POS',-999.,struct[0],infile,logfile)
	    fo_volts = appendkey(fo_volts,'FO-VOLTS',-999.,struct[0],infile,logfile)
	    fo_pos_s = appendkey(fo_pos_s,'FO-POS-S',-999.,struct[0],infile,logfile)
	    fo_pos_v = appendkey(fo_pos_v,'FO-POS-V',-999.,struct[0],infile,logfile)
	    gr_state = appendkey(gr_state,'GR-STATE','UNKNOWN',struct[0],infile,logfile)
	    gr_sta = appendkey(gr_sta,'GR-STA','UNKNOWN',struct[0],infile,logfile)
	    gr_angle = appendkey(gr_angle,'GR-ANGLE',-999.,struct[0],infile,logfile)
	    gm_steps = appendkey(gm_steps,'GM-STEPS',-999,struct[0],infile,logfile)
	    gm_volts = appendkey(gm_volts,'GM-VOLTS',-999.,struct[0],infile,logfile)
	    gr_sta_s = appendkey(gr_sta_s,'GR-STA-S',-999.,struct[0],infile,logfile)
	    gr_sta_v = appendkey(gr_sta_v,'GR-STA-V',-999.,struct[0],infile,logfile)
	    gr_steps = appendkey(gr_steps,'GR-STEPS',-999,struct[0],infile,logfile)
	    grating = appendkey(grating,'GRATING','UNKNOWN',struct[0],infile,logfile)
	    grtilt = appendkey(grtilt,'GRTILT',-999.,struct[0],infile,logfile)
	    bs_state = appendkey(bs_state,'BS-STATE','UNKNOWN',struct[0],infile,logfile)
	    fi_state = appendkey(fi_state,'FI-STATE','UNKNOWN',struct[0],infile,logfile)
            fi_sta = appendkey(fi_sta,'FI-STA','UNKNOWN',struct[0],infile,logfile)
	    fm_steps = appendkey(fm_steps,'FM-STEPS',-999,struct[0],infile,logfile)
	    fm_volts = appendkey(fm_volts,'FM-VOLTS',-999.,struct[0],infile,logfile)
	    fm_sta_s = appendkey(fm_sta_s,'FM-STA-S',-999.,struct[0],infile,logfile)
	    fm_sta_v = appendkey(fm_sta_v,'FM-STA-V',-999.,struct[0],infile,logfile)
	    ar_state = appendkey(ar_state,'AR-STATE','UNKNOWN',struct[0],infile,logfile)
	    ar_sta = appendkey(ar_sta,'AR-STA','UNKNOWN',struct[0],infile,logfile)
	    camang = appendkey(camang,'CAMANG',-999.,struct[0],infile,logfile)
	    ar_sta_s = appendkey(ar_sta_s,'AR-STA-S',-999.,struct[0],infile,logfile)
	    ar_angle = appendkey(ar_angle,'AR-ANGLE',-999.,struct[0],infile,logfile)
	    coltemp = appendkey(coltemp,'COLTEMP',-999.,struct[0],infile,logfile)
	    camtemp = appendkey(camtemp,'CAMTEMP',-999.,struct[0],infile,logfile)
	    proc = appendkey(proc,'PROC','UNKNOWN',struct[0],infile,logfile)

# SALTICAM-only keywords

	if (status == 0 and scam): 
	    filpos = appendkey(filpos,'FILPOS',-999,struct[0],infile,logfile)

# close image files

	if (status == 0): status = saltio.closefits(struct,logfile)

# progress log statement

	if (status == 0): saltprint.log(logfile,'SALTLOG -- read ' + infile,verbose)

# delete output file if it exists

    if (status == 0 and clobber and os.path.isfile(outfile)): 
	status = saltio.delete(outfile,False,logfile)

# define generic columns of output table

    if (status == 0):
	col.append(pyfits.Column(name='FILENAME',format='32A',array=filename))
	col.append(pyfits.Column(name='PROPID',format='20A',array=propid))
	col.append(pyfits.Column(name='PROPOSER',format='20A',array=proposer))
	col.append(pyfits.Column(name='OBJECT',format='16A',array=object))
	col.append(pyfits.Column(name='RA',format='11A',unit='hours',array=ra))
	col.append(pyfits.Column(name='DEC',format='12A',unit='deg',array=dec))
	col.append(pyfits.Column(name='OBJEPOCH',format='E',unit='year',array=objepoch))
	col.append(pyfits.Column(name='EQUINOX',format='E',unit='year',array=equinox))
	col.append(pyfits.Column(name='DATE-OBS',format='10A',array=date_obs))
	col.append(pyfits.Column(name='UTC-OBS',format='12A',unit='hours',array=utc_obs))
	col.append(pyfits.Column(name='TIME-OBS',format='12A',unit='hours',array=time_obs))
	col.append(pyfits.Column(name='EXPTIME',format='D',unit='s',array=exptime))
	col.append(pyfits.Column(name='OBSMODE',format='16A',array=obsmode))
	col.append(pyfits.Column(name='DETMODE',format='20A',array=detmode))
	col.append(pyfits.Column(name='CCDTYPE',format='8A',array=ccdtype))
	col.append(pyfits.Column(name='DETSIZE',format='23A',unit='pixels',array=detsize))
	col.append(pyfits.Column(name='NCCDS',format='I',array=nccds))
	col.append(pyfits.Column(name='CCDSUM',format='5A',unit='pixels',array=ccdsum))
	col.append(pyfits.Column(name='GAINSET',format='6A',array=gainset))
	col.append(pyfits.Column(name='ROSPEED',format='4A',array=rospeed))
	col.append(pyfits.Column(name='INSTRUME',format='8A',array=instrume))
	col.append(pyfits.Column(name='FILTER',format='8A',array=filter))
	col.append(pyfits.Column(name='FOCUS',format='J',unit='steps',array=focus))
	col.append(pyfits.Column(name='TELHA',format='11A',unit='deg',array=telha))
	col.append(pyfits.Column(name='TELRA',format='11A',unit='hours',array=telra))
	col.append(pyfits.Column(name='TELDEC',format='12A',unit='deg',array=teldec))
	col.append(pyfits.Column(name='TELPA',format='E',unit='deg',array=telpa))
	col.append(pyfits.Column(name='TELAZ',format='E',unit='deg',array=telaz))
	col.append(pyfits.Column(name='TELALT',format='E',unit='deg',array=telalt))
	col.append(pyfits.Column(name='TRKX',format='E',unit='m',array=trkx))
	col.append(pyfits.Column(name='TRKY',format='E',unit='m',array=trky))
	col.append(pyfits.Column(name='TRKZ',format='E',unit='m',array=trkz))
	col.append(pyfits.Column(name='TRKPHI',format='E',unit='arcsec',array=trkphi))
	col.append(pyfits.Column(name='TRKTHETA',format='E',unit='arcsec',array=trktheta))
	col.append(pyfits.Column(name='TRKRHO',format='E',unit='deg',array=trkrho))
	col.append(pyfits.Column(name='INTFR',format='E',unit='m',array=intfr))
	col.append(pyfits.Column(name='COLPHI',format='E',unit='arcsec',array=colphi))
	col.append(pyfits.Column(name='COLTHETA',format='E',unit='arcsec',array=coltheta))
	col.append(pyfits.Column(name='TELTEMP',format='E',unit='C',array=teltemp))
	col.append(pyfits.Column(name='PAYLTEMP',format='E',unit='C',array=payltemp))
	col.append(pyfits.Column(name='CCDTEM',format='D',unit='K',array=ccdtem))
	col.append(pyfits.Column(name='DEWTEM',format='E',unit='C',array=dewtem))
	col.append(pyfits.Column(name='AMPTEM',format='E',unit='C',array=amptem))
	col.append(pyfits.Column(name='CENTEM',format='D',unit='C',array=centem))
	col.append(pyfits.Column(name='DETSWV',format='16A',array=detswv))

# define RSS columns of output table

    if (status == 0 and rss):
	col.append(pyfits.Column(name='DEWPRE',format='D',array=dewpre))
	col.append(pyfits.Column(name='POSANG',format='E',unit='deg',array=posang))
	col.append(pyfits.Column(name='LAMPID',format='8A',array=lampid))
	col.append(pyfits.Column(name='CALFILT',format='8A',array=calfilt))
	col.append(pyfits.Column(name='CALND',format='E',array=calnd))
	col.append(pyfits.Column(name='TELRHO',format='E',unit='deg',array=telrho))
	col.append(pyfits.Column(name='PELLICLE',format='8A',array=pellicle))
	col.append(pyfits.Column(name='INSTPORT',format='8A',array=instport))
	col.append(pyfits.Column(name='CF-STATE',format='20A',array=cf_state))
	col.append(pyfits.Column(name='SM-STATE',format='20A',array=sm_state))
	col.append(pyfits.Column(name='SM-STA',format='8A',array=sm_sta))
	col.append(pyfits.Column(name='SM-STEPS',format='J',unit='steps',array=sm_steps))
	col.append(pyfits.Column(name='SM-VOLTS',format='E',unit='volts',array=sm_volts))
	col.append(pyfits.Column(name='SM-STA-S',format='E',array=sm_sta_s))
	col.append(pyfits.Column(name='SM-STA-V',format='E',array=sm_sta_v))
	col.append(pyfits.Column(name='MASKID',format='16A',array=maskid))
	col.append(pyfits.Column(name='MASKTYP',format='16A',array=masktyp))
	col.append(pyfits.Column(name='WP-STATE',format='20A',array=wp_state))
	col.append(pyfits.Column(name='HWP-CMD',format='16A',array=hwp_cmd))
	col.append(pyfits.Column(name='HW-STEPS',format='J',unit='steps',array=hw_steps))
	col.append(pyfits.Column(name='HWP-STA',format='E',array=hwp_sta))
	col.append(pyfits.Column(name='QWP-CMD',format='16A',array=qwp_cmd))
	col.append(pyfits.Column(name='QW-STEPS',format='J',unit='steps',array=qw_steps))
	col.append(pyfits.Column(name='QWP-STA',format='E',array=qwp_sta))
	col.append(pyfits.Column(name='QWP-ANG',format='E',unit='deg',array=qwp_ang))
	col.append(pyfits.Column(name='HWP-ANG',format='E',unit='deg',array=hwp_ang))
	col.append(pyfits.Column(name='SH-STATE',format='20A',array=sh_state))
	col.append(pyfits.Column(name='FO-STATE',format='20A',array=fo_state))
	col.append(pyfits.Column(name='FO-POS',format='E',array=fo_pos))
	col.append(pyfits.Column(name='FO-VOLTS',format='E',unit='volts',array=fo_volts))
	col.append(pyfits.Column(name='FO-POS-S',format='E',array=fo_pos_s))
	col.append(pyfits.Column(name='FO-POS-V',format='E',array=fo_pos_v))
	col.append(pyfits.Column(name='GR-STATE',format='20A',array=gr_state))
	col.append(pyfits.Column(name='GR-STA',format='10A',array=gr_sta))
	col.append(pyfits.Column(name='GR-ANGLE',format='E',unit='deg',array=gr_angle))
	col.append(pyfits.Column(name='GM-STEPS',format='J',unit='steps',array=gm_steps))
	col.append(pyfits.Column(name='GM-VOLTS',format='E',unit='volts',array=gm_volts))
	col.append(pyfits.Column(name='GR-STA-S',format='E',array=gr_sta_s))
	col.append(pyfits.Column(name='GR-STA-V',format='E',array=gr_sta_v))
	col.append(pyfits.Column(name='GR-STEPS',format='J',unit='steps',array=gr_steps))
	col.append(pyfits.Column(name='GRATING',format='8A',array=grating))
	col.append(pyfits.Column(name='GRTILT',format='E',unit='deg',array=grtilt))
	col.append(pyfits.Column(name='BS-STATE',format='24A',array=bs_state))
	col.append(pyfits.Column(name='FI-STATE',format='20A',array=fi_state))
        col.append(pyfits.Column(name='FI-STA',format='7A',array=fi_sta))
	col.append(pyfits.Column(name='FM-STEPS',format='J',unit='steps',array=fm_steps))
	col.append(pyfits.Column(name='FM-VOLTS',format='E',unit='volts',array=fm_volts))
	col.append(pyfits.Column(name='FM-STA-S',format='E',array=fm_sta_s))
	col.append(pyfits.Column(name='FM-STA-V',format='E',array=fm_sta_v))
	col.append(pyfits.Column(name='AR-STATE',format='24A',array=ar_state))
	col.append(pyfits.Column(name='AR-STA',format='16A',array=ar_sta))
	col.append(pyfits.Column(name='CAMANG',format='E',unit='deg',array=camang))
	col.append(pyfits.Column(name='AR-STA-S',format='E',array=ar_sta_s))
	col.append(pyfits.Column(name='AR-ANGLE',format='E',unit='deg',array=ar_angle))
	col.append(pyfits.Column(name='COLTEMP',format='E',unit='C',array=coltemp))
	col.append(pyfits.Column(name='CAMTEMP',format='E',unit='C',array=camtemp))
	col.append(pyfits.Column(name='PROC',format='20A',array=proc))
	col.append(pyfits.Column(name='PCS-VER',format='4A',array=pcs_ver))

# define SALTICAM columns of output table

    if (status == 0 and scam):
	col.append(pyfits.Column(name='FILPOS',format='I',array=filpos))

# construct FITS table from columns

    if (status == 0): table,status = saltio.fitscolumns(col,outfile,logfile)

# write FITS table to output file

    if (status == 0): struct,status = saltio.newfitstable(table,outfile,logfile)

# name the table extension

    if (status == 0): status = saltkey.new('EXTNAME','OBSLOG','extension name',
					   struct,outfile,logfile)

# housekeeping keywords

    if (status == 0): status = saltkey.put('SAL-TLM',time.asctime(time.localtime()),
					   struct,outfile,logfile)
    if (status == 0): status = saltkey.history(struct,history1,outfile,logfile)

# close table file

    message = '\nSALTLOG -- created observation log ' + outfile
    if (status == 0): status = saltio.writefits(struct,outfile,logfile)
    if (status == 0): saltprint.log(logfile,message,verbose)

# end time

    if (status == 0):
	saltprint.time('\nSALTLOG -- completed at',logfile,verbose)
    else:
	saltprint.time('\nSALTLOG -- aborted at',logfile,verbose)

# -----------------------------------------------------------
# read keyword and append to list 

def appendkey(list,keyword,default,struct,infile,logfile):

    try:
	value = struct.header[keyword]
    except:
	value = default
	message = 'WARNING -- SALTLOG.APPENDKEY: cannot find keyword ' + infile + '[0][' + keyword + ']'
	status = saltprint.err(logfile,message)
    if (str(value).strip() == ''): value = default
    list.append(str(value).lstrip().rstrip())

    return list


# -----------------------------------------------------------
# main code 

parfile = iraf.osfn("pipetools$saltlog.par") 
t = iraf.IrafTaskFactory(taskname="saltlog",value=parfile,function=saltlog, pkgname='pipetools')
